/* ****************************************** */
/* Emu75: HP-75  software emulator for PC     */
/*                                            */
/* monit.c   monitor module for Emu75         */
/*                                            */
/* Version 1.3  J-F Garnier 2010-2018         */
/* ****************************************** */

#include <stdio.h>
#include <string.h>
#include <dos.h>
#include <conio.h>
#include <time.h>

#include "trans.h"

#define GLOBAL extern
#include "capricor.h"
#include "hpil.h"

/* variables for breakpoints */
static unsigned char lastopcode=0;
static unsigned int lastaddr=0;

static unsigned int regPC, regSP;   // values from CPU registers


extern int desas(int addr, char *codes, char *ligne); /* desas75.c */
extern long ch_symb(char *s);   /* trans.c */




/* ****************************************** */
/* hextoint()                                 */
/*                                            */
/* convert a hex nibble into a int            */
/* ****************************************** */
int hextoint(char c)
{
  if (c>'9') c -= 7;
  c -= 48;
  return(c);
}




/* ****************************************** */
/* inttohex()                                 */
/*                                            */
/* convert a int nibble (0-15) into a hex char*/
/* ****************************************** */
char inttohex(int c)
{
  c &= 15;
  if (c>9) c += 7;
  return(c+48);
}




/* ****************************************** */
/* lit()                                      */
/*                                            */
/* read n bytes at address adr                */
/* ****************************************** */
void lit(unsigned int adr, int n, char *buf)
{
  char far *pt;
  int i;
  int adrh;
  int segment, offset;

  adrh= (adr>>13);
  adrh &= 0x07;
  segment=tab_seg[adrh];
  offset = adr & 0x1fff;
  if (segment==0) {
    for (i=0;i<n;i++) *buf++ = 0;
  }
  else {
    pt=MK_FP(segment,offset);
    for (i=1;i<=n;i++) {
      *buf++ = *pt++;
    }
  }
}

int lit1(unsigned int adr)
{
  unsigned char buf[2];

  lit(adr,1,buf);
  return (buf[0]);
}

int lit2(unsigned int adr)
{
  unsigned char buf[2];

  lit(adr,2,buf);
  return (buf[0]+(buf[1]<<8));
}



void alpha(char *dest,int n)
{
  int i, c;

  for (i=0;i<n;i++) {
    c = *dest;
    c &= 127;   // remove underlined attribute
    if (c==0) break;
    if (c==8) c=126;   // remap the HP75 char ~
    if (c==10) c=127;  // remap the HP75 char 
    if ((c<32)||(c>127)) c = '.';
    *dest++ =c;
  }
}






/* display desassembled opcodes */
int dasm1(unsigned int adr)
{
  int j, n;
  unsigned char codes[24];
  char ligne[80];

  lit(adr,22,codes);
  n=desas(adr,codes,ligne);
  printf(" %04X ",adr);
  for (j=0;(j<n)&&(j<3);j++) printf("%02X ",codes[j]);
  for (;j<3;j++) printf("   ");
  printf(" %s\n",ligne);
  adr += n;
  return(adr);
}









void affiche(int full)
{
  int i, j,  n;
  int v;

  // set regPC and regSP:
  regPC=regCPU.reg.pc;
  regSP=regCPU.reg.sp;


  if (full) {
/*
R00: 00 00 00 00 00 00 00 00    R40: 00 00 00 00 00 00 00 00
R10: 00 00 00 00 00 00 00 00    R50: 00 00 00 00 00 00 00 00
R20: 00 00 00 00 00 00 00 00    R60: 00 00 00 00 00 00 00 00
R30: 00 00 00 00 00 00 00 00    R70: 00 00 00 00 00 00 00 00
*/
    for (i=0;i<4;i++) {
      printf("R%02o=",i*8);
      for (j=0;j<8;j+=2)
        printf(" %02x:%02x",regCPU.byt[i*8+j],regCPU.byt[i*8+j+1]);
      printf("   R%02o=",32+i*8);
      for (j=0;j<8;j+=2)
        printf(" %02x:%02x",regCPU.byt[32+i*8+j],regCPU.byt[32+i*8+j+1]);
      printf("\n");
    }

  }
  else {
    /* display only the 8-byte reg pointed by ARP and DRP */
    i=regDRP&0x38;
    printf("R%02o=",i);
    for (j=0;j<8;j+=2)
      printf(" %02x:%02x",regCPU.byt[i+j],regCPU.byt[i+j+1]);
    i=regARP&0x38;
    printf("   R%02o=",i);
    for (j=0;j<8;j+=2)
      printf(" %02x:%02x",regCPU.byt[i+j],regCPU.byt[i+j+1]);
    printf("\n");
  }

  printf("PC=%04x SP=%04x  ROMSLT=%02x     ",regPC, regSP, regROMslt);
  printf("DRP=%02o ARP=%02o E=%01X   GIE=%d\n",regDRP, regARP, regE, fintr);
  n=regFlags;
  printf("Flags: MSB=%d LDZ=%d OVF=%d DCM=%d",(n>>7)&1, (n>>6)&1, (n>>5)&1, (n>>4)&1);
  printf(" CY=%d Z=%d RDZ=%d LSB=%d\n", (n>>3)&1, (n>>2)&1, (n>>1)&1, n&1);
  printf("Stack:");
  for (i=0;i<16;i++) {
    if ( ((regSP-1-i)&0xc000) == 0x8000) {
      v=lit1(regSP-1-i);
      printf(" %02x", v );
    }
  }
  printf("\n");
  dasm1(regPC);
}


void affiche_stack(void)
{
  int i, j;
  int v;

  printf("Stack:\n");
  for (j=0;j<64;j+=16) {
    for (i=0;i<16;i++) {
      if ( ((regSP-1-j-i)&0xc000) == 0x8000) {
        v=lit1(regSP-1-j-i);
        printf(" %02x", v );
      }
    }
    printf("\n");
  }
}


void set_brkpt(unsigned adr)
{
  unsigned char far *ptr;
  int adrh, segment, offset;

  if (adr>0) {
    adrh= (adr>>13);
    adrh &= 0x07;
    segment=tab_seg[adrh];
    offset = adr & 0x1fff;
    if (segment!=0) {
      lastaddr=adr;
      ptr=MK_FP(segment,offset);
      lastopcode=*ptr;
      *ptr=0xde;  // use NOP1 unused opcode
    }
  }
}


void res_brkpt(void)
{
  unsigned char far *ptr;
  int adrh, segment, offset;

  if (lastaddr>0) {
    adrh= (lastaddr>>13);
    adrh &= 0x07;
    segment=tab_seg[adrh];
    offset = lastaddr & 0x1fff;
    if (segment!=0) {
      ptr=MK_FP(segment,offset);
      if ((*ptr)==0xde) *ptr=lastopcode;  // security for switched ROMs
    }
    lastaddr=0;
  }
}







int monitor(int sst)
{
  int n;

  n=executeCapricorn(sst);
  if (n==-1)
    printf("invalid instruction\n");
  return(n);
}


/* display buffer */
/*
*/
void view_display(void)
{
  char ligne[82];

  lit(0x82e6,78,ligne);
  ligne[78]=0;
  alpha(ligne,78);
  printf("%s\n",ligne);
}



void view_xmem(void)
{
  unsigned char far *ptr;
  int i;

  printf("Installed MEMIC=%d  RAM=%dkB",instmemic,xmemsize);
  printf(" - selected MEMIC=%d\n",podmemic);
  printf("reg A=%02x%02x%02x  ",podregA[2],podregA[1],podregA[0]);
  printf("B=%02x%02x%02x  ",podregB[2],podregB[1],podregB[0]);
  printf("C=%02x%02x%02x  ",podregC[2],podregC[1],podregC[0]);
  printf("D=%02x%02x%02x\n",podregD[2],podregD[1],podregD[0]);
  printf("stat= %02x %02x %02x %02x\n",podsts[0],podsts[1],podsts[2],podsts[3]);
  if (xmem_seg[0]!=0) {
    printf("IMEM:");
    ptr= (unsigned char far *) MK_FP(xmem_seg[0],0);
    for (i=0;i<16;i++) printf(" %02x",*ptr++);
    printf("\n");
  }
  if (xmem_seg[4]!=0) {
    printf("XMEM:");
    ptr= (unsigned char far *) MK_FP(xmem_seg[4],0);
    for (i=0;i<16;i++) printf(" %02x",*ptr++);
    printf("\n");
  }
}



void help(void)
{
  printf("a slot        : activate ROM slot\n");
  printf("c             : check ROM\n");
  printf("d [ . | addr ]: disasm (.=current PC)\n");
  printf("g addr        : go, run until PC=addr\n");
  printf("j addr        : jump, set PC=addr\n");
  printf("+             : increment PC\n");
  printf("k x           : push keycode x into buffer\n");
//  printf("l             : load processor state and RAMs\n");
  printf("m [ . | addr] : memory dump\n");
  printf("q             : quit\n");
  printf("r [ addr ]    : run emulator [break at addr]\n");
//  printf("s             : save processor state and RAMs\n");
  printf("t             : trace (single step)\n");
  printf("vd            : view display buffer\n");
  printf("vl            : view LCD\n");
  printf("vp            : view PIL registers (HPIL)\n");
  printf("vr            : view CPU registers\n");
  printf("vs            : view extended stack\n");
  printf("vt            : view timer registers\n");
  printf("vx            : view xmem registers\n");
//  printf("x             : reset all: ROM, RAM, registers\n");

}




/* returns 1 if deep sleep, and -1 for command 'q' */
int commande(void)
{
  char cmd[82];
  char ligne[82];
  char *tok;
  unsigned char codes[64], c;
  unsigned int debut, fin;
  int  ret;
  unsigned int addr;
  int i, j;

  fin_ecran();
  printf("Emu75 Ready.\n");
  setad(regARP,regDRP);
  affiche(1);
  fin=0;
  do {
    ret=0;
    printf("Cmd: ");
    gets(cmd);
    tok=strtok(cmd," ");
    tok=strtok(NULL," ");
    if (tok!=NULL) {
      if (tok[0]=='.') {
	debut=regPC;
        setad(regARP,regDRP);
      }
      else if (tok[0]=='=') {
	addr=ch_symb(tok);
	if (addr!= -1)
	  debut=addr;
	else {
	  printf("%s not found\n",tok);
	  cmd[0]=0;
        }
      }
      else
	sscanf(tok,"%x",&debut);
    }
    else
      debut=fin;

    switch (cmd[0]) {

      case 'a':
        selectbank(debut);
        break;

      case 'c':
	checkrom();
	break;

      case 'd':
	addr=debut;
	for (i=0;i<20;i++) {
          if (ch_label(addr,ligne)) {
	    printf("=%s\n",ligne);
            i++;
          }
	  addr=dasm1(addr);
	}
	fin=addr;
	break;

      case 'g':
	if (tok!=NULL) set_brkpt(debut);
//	init_timer();
	do {
	  ret=monitor(1);
	  if (kbhit()) {
            c=getch();
	    if (c==27) ret=-2;
	  }
	} while (ret==0);
	res_brkpt();
        setad(regARP,regDRP);
	affiche(0);
	ret=0;
	break;

      case 'j':
	regPC=debut;
        regCPU.reg.pc=regPC;
        setad(regARP,regDRP);
	affiche(0);
	break;

      case '+':
        regPC++;
        regCPU.reg.pc=regPC;
        setad(regARP,regDRP);
	affiche(0);
        break;

      case 'k':
	push_key(debut);
	break;

      case 'l':
	load_state();
	break;

      case 'm':
	fin=debut+256;
        addr=debut;
	for (j=0;j<16;j++) {
	  lit(addr,16,codes);
	  printf("%04x ",addr);
          for (i=0;i<8;i++)
	    printf("%02X ",codes[i]);
	  printf("- ");
          for (;i<16;i++)
	    printf("%02X ",codes[i]);
	  printf("  ");
          for (i=0;i<16;i++) {
            c=codes[i];
            if ((c<32)||(c>127)) c='.';
	    printf("%c",c);
          }
	  printf("\n");
          addr +=16;
	}
	break;

      case 'q':
	ret=-1;
	break;

      case 'r':
	if (tok!=NULL) set_brkpt(debut);
        prepa_ecran(0);
	ret=simulator();  /* returns 1 if deep sleep */
        fin_ecran();
	res_brkpt();
	if (ret!=1) {
          ret=0;
          setad(regARP,regDRP);
          affiche(1);
        }
	break;

      case 's':
	save_state();
	break;

      case 't':
	do {
//	  init_timer();
	  monitor(1);
          setad(regARP,regDRP);
	  affiche(0);
	} while (getch()!=27);
	break;

      case 'v':
	switch (cmd[1]) {
	  case 'd':
	    view_display();
	    break;
          case 'l':
            strncpy(ligne,regLCD,32);
            ligne[32]=0;
            printf("%s\n\r",ligne);
            break;
          case 'p':
            printf("hpil reg: ");
            for (i=0;i<8;i++) printf(" %02x",hpil_reg[i]);
            printf("  (%02x)\n",(hpil_reg[8]&0xe0)|flgenb);
            break;
          case 'r':
	    affiche(1);
	    break;
	  case 's':
	    affiche_stack();
	    break;
	  case 't':
            i=lit1(0xff81);
            printf("Timer = %08lx%02x , stat=%02x\n",timer,timer_l,i);
            i=lit1(0xff80);
            printf("Comp  = %08lx%02x , stat=%02x\n",tcomp,tcomp_l,i);
	    break;
          case 'x':
            view_xmem();
            break;
	  default:
	    printf("?\n");
	}
	break;

      case 'x':
	init_register();
	loadrom();
	init_ram();
	break;

      case '?':
	help();
	break;

      default:
	printf("?\n");
    }
  } while (ret==0);

  return(ret);
}
